import numpy as np

JC = 'JC'
F81 = 'F81'
EFT = 'EFT'


def is_f81_like(model):
    """
    Checks if the evolutionary model is F81 or its simplification, e.g. JC or EFT.

    :param model: evolutionary model
    :type model: str
    :rtype: bool
    """
    return model in {F81, JC, EFT}


def get_mu(frequencies):
    """
    Calculates the mutation rate for F81 (and JC that is a simplification of it),
    as \mu = 1 / (1 - sum_i \pi_i^2). This way the overall rate of mutation -\mu trace(\Pi Q) is 1.
    See [Gascuel "Mathematics of Evolution and Phylogeny" 2005] for further details.

    :param frequencies: numpy array of state frequencies \pi_i
    :return: mutation rate \mu = 1 / (1 - sum_i \pi_i^2)
    """
    return 1. / (1. - frequencies.dot(frequencies))


def get_f81_pij(t, frequencies, mu):
    """
    Calculate the probability of substitution i->j over time t, given the mutation rate mu:
    For F81 (and JC which is a simpler version of it)
    Pij(t) = \pi_j (1 - exp(-mu t)) + exp(-mu t), if i == j, \pi_j (1 - exp(-mu t)), otherwise
    [Gascuel "Mathematics of Evolution and Phylogeny" 2005].

    :param frequencies: array of state frequencies \pi_i
    :type frequencies: numpy.array
    :param mu: mutation rate: \mu = 1 / (1 - sum_i \pi_i^2)
    :type mu: float
    :param t: time t
    :type t: float
    :param sf: scaling factor by which t should be multiplied.
    :type sf: float
    :return: probability matrix
    :rtype: numpy.ndarray
    """

    # if mu == inf (e.g. just one state) and t == 0, we should prioritise mu
    exp_mu_t = 0. if (mu == np.inf) else np.exp(-mu * t)
    return (1 - exp_mu_t) * frequencies + np.eye(len(frequencies)) * exp_mu_t



from collections import Counter

import numpy as np

from pastml.annotation import get_forest_stats
from pastml.ml import get_pij_method


def simulate_states(tree, model, frequencies, kappa, tau, sf, character, rate_matrix=None, n_repetitions=1_000):
    n_states = len(frequencies)
    state_ids = np.array(range(n_states))

    avg_br_len, num_nodes, num_tips, tree_len = get_forest_stats([tree])
    num_edges = num_nodes - 1
    tau_factor = tree_len / (tree_len + tau * num_edges)
    get_pij = get_pij_method(model, frequencies, kappa, rate_matrix=rate_matrix)

    for n in tree.traverse('levelorder'):
        if n.is_root():
            random_states = np.random.choice(state_ids, size=n_repetitions, p=frequencies)
        else:
            probs = get_pij((n.dist + tau) * tau_factor * sf)
            probs = np.maximum(probs, 0)
            random_states = np.zeros(n_repetitions, dtype=int)
            parent_states = getattr(n.up, character)
            sorted_indices = np.argsort(parent_states)
            parent_state_nums = Counter(parent_states)
            offset = 0
            for i in state_ids:
                parent_nums_i = parent_state_nums[i]
                if parent_nums_i > 0:
                    random_states[sorted_indices[offset: offset + parent_nums_i]] = \
                        np.random.choice(state_ids, size=parent_nums_i, p=probs[i])
                    offset += parent_nums_i
        n.add_feature(character, random_states)

    return tree
